# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""Observatory is used to change behavior on some of the structures and components
   in the Scheduler service and others.

   Right now only Gemini Observatory is implemented but more should come.
   For new implementations, see Abstract.
"""
